/*
 *        x
 *      -----> 
 *      0 1 2 3 4 5 6 7
 *  y  0 | | |0| | | | | |
 *  |  1 | | | | | |0| | |
 *  |  2 | | | | | | | |0|
 * \/  3 |0| | | | | | | |
 *     4 | | | | |0| | | |
 *     5 | | | | | | |0| |
 *     6 | |0| | | | | | |
 *     7 | | | |0| | | | |
 */


import java.util.Random;
import java.util.*;

/**
 * Class Sachovnice represents chessboard and solves Eight queen puzzle by means of hill-climbing algorithm.
 * 
 * @author Martin ehnek, 143357@mail.muni.cz
 * @version 
 */
public class Sachovnice
{
    private Pozice[] damy;
    private int[][] bitmap = new int[8][8];  //pole poli, tam kde nic neni je 0 a tam kde je dama je 1 
    
    
    /**
	 * Constructor for objects of class Sachovnice.
	 * @param damy Array of objects of class Pozice that represent positions of queens.
	 */
	public Sachovnice(Pozice[] damy)
    {
        this.damy = damy;
        for(int i = 0; i<8; i++)
          bitmap[damy[i].getX()][damy[i].getY()] = 1;
    }
    
    
    /**
	 * Nonparametric constructor for objects of class Sachovnice that creates random chessboard.
	 */
	public Sachovnice() 
    {
       int x = 0;
       int y = 0;
       damy = new Pozice[8];
       Random mujRandom = new Random();
       for(int i = 0; i<8; i++) //pro vsech 8 dam
         while (true)
          { y = mujRandom.nextInt(8);
            if (bitmap[x][y] == 0) 
              {
                  bitmap[x][y] = 1;
                  damy[i] = new Pozice(x,y);
                  x++;
                  break;
               }
           }
     }
    
    
    /**
     * Returns a string representation of the object.
     * @return A string representation of the object.
     */
    public String toString()  //chtel jsem to udelat Unicodem, at to vypada dobre, ale bohuzel
    {
        boolean barva = true;  //true bila, false cerna
        int i = 0;
        String navrat = "";
        for (i = 0; i<8; i++)
          {   //navrat+="-----------------\n";
              navrat+='|';
              for (int j = 0; j<8; j++)
                {                                       //prazdne policko     //policko s damou  
                    navrat+= (bitmap[j][i] == 0) ? (barva ? ' ' : '#') : (barva ? '0' : 'o');
                      navrat += '|';
                    barva = !barva;  
                 }
              navrat+='\n';
              barva = !barva;
           }
        //navrat+="-----------------\n";   
        return navrat;
     }     
    
    
     /**
      * Returns a number of conflicts with other queens that has the queen in paramether
      * @return A number of conflicts with other queens that has the queen in paramether
      * @param dama Chosen queen.
      */
     private int konfliktu(Pozice dama)  
     {
         if (dama == null)
          throw new NullPointerException("dama");
         return konfliktu(dama.getX(), dama.getY());
     }
     
     
     
     /**
      * Returns a number of conflicts with other queens that has the queen at chosen coordinates.
      * @return A number of conflicts with other queens that has the queen at chosen coordinates.
      * @param x X coordinate of the queen.
      * @param y Y coordinate of the queen.
      */
     private int konfliktu(int x, int y)  //prozkoumej nahoru dolu, doprava doleva, diagonalne, jestli neni konflikt 
     {   
          int j = 0;
          int k = 0;
          int pocet = 0;
          for(j = 0; j<8; j++)  //prozkoumat vodorovne
             {
                 if ( (j != x) && (bitmap[j][y] == 1)  )
                  {
                      pocet++; //System.out.print("V,");
                  }
                   
              }
           
          int newX = 0;
          int newY = 0;
          for(j = 1; j<7; j++) //prozkoumat diagonalne //mohlo by to byt v cyklu(ech), tim by ale dost klesla srozumitelnost
             {  
                // [+j][+j] 
               if ( ( (newX = x + j) >= 0) && (newX<8) && ( (newY = y + j) >= 0 ) && (newY<8) )   
                 if ((newX != x) && (newY != y) && (bitmap[newX][newY] == 1))
                   {pocet++; //System.out.print("D1,");
                    }
               // [+j][-j] 
               if ( ( (newX = x + j) >= 0) && (newX<8) && ( (newY = y - j) >= 0 ) && (newY<8) ) 
                 if ((newX != x) && (newY != y) && (bitmap[newX][newY] == 1))
                   {pocet++; //System.out.print("D2,");
                    }
               // [-j][+j] 
               if ( ( (newX = x - j) >= 0) && (newX<8) && ( (newY = y + j) >= 0 ) && (newY<8) ) 
                 if ((newX != x) && (newY != y) && (bitmap[newX][newY] == 1))
                   {pocet++; //System.out.print("D3,");
                   }
               // [-j][-j] 
               if ( ( (newX = x - j) >= 0) && (newX<8) && ( (newY = y - j) >= 0 ) && (newY<8) ) 
                 if ((newX != x) && (newY != y) && (bitmap[newX][newY] == 1))
                   {pocet++; //System.out.print("D4,");
                    }
             } 
          return pocet ;   
     }
     
     
      
      /**
       * Solves two chosen chessboard setting and one random setting.
       */
      public static void main()
       {
         Pozice[] prednaskaDamy1 = {new Pozice(0,1),new Pozice(1,4),new Pozice(2,6),new Pozice(3,3),new Pozice(4,0),new Pozice(5,2),new Pozice(6,5),new Pozice(7,7) } ;
         Pozice[] prednaskaDamy2 = {new Pozice(0,0),new Pozice(1,1),new Pozice(2,2),new Pozice(3,3),new Pozice(4,4),new Pozice(5,5),new Pozice(6,6),new Pozice(7,7) } ;
         Pozice[] prednaskaDamy3 = {new Pozice(0,4),new Pozice(1,5),new Pozice(2,6),new Pozice(3,3),new Pozice(4,4),new Pozice(5,5),new Pozice(6,6),new Pozice(7,5) } ;
         Pozice[] spatna = {new Pozice(0,1),new Pozice(1,4),new Pozice(2,4),new Pozice(3,2),new Pozice(4,6),new Pozice(5,0),new Pozice(6,2),new Pozice(7,6) } ;
         //Pozice[] damy1 = {new Pozice(0,7),new Pozice(1,4),new Pozice(2,3),new Pozice(3,6),new Pozice(4,5),new Pozice(5,0),new Pozice(6,4),new Pozice(7,6) } ;
         //Pozice[] damy2 = {new Pozice(0,6),new Pozice(1,4),new Pozice(2,0),new Pozice(3,0),new Pozice(4,3),new Pozice(5,7),new Pozice(6,0),new Pozice(7,1) } ;
         
         String separator = "========================================\n";
        /* Sachovnice easy = new Sachovnice(prednaskaDamy1);
         System.out.println("\nPrvni sachovice z prednasky (uloha 3):\n" + easy);
         easy.Vyres(true, 1000, 3); //pro zobrazeni vypisu prepsat na true 
         System.out.println(separator);
         
         Sachovnice diagonala = new Sachovnice(prednaskaDamy2);
         System.out.println("\nDruha sachovnice z prednasky (diagonala):\n" + diagonala );
         diagonala.Vyres(false, 1000, 3); 
         System.out.println(separator);
         
         Sachovnice mocKonfliktu = new Sachovnice(prednaskaDamy3);
         System.out.println("\nSachovnice ze zadani s hodne konflikty :\n" + mocKonfliktu );
         mocKonfliktu.Vyres(false, 1000, 3); 
         System.out.println(separator);
         
         Sachovnice nahoda = new Sachovnice(); //konstruktor bez paramatru vygeneruje nahodnou sachovnici
         System.out.println("\nNahodne vygenerovana sachovnice:\n" + nahoda);
         nahoda.Vyres(false, 1000, 3); 
         System.out.println(separator);
         */
         
         Sachovnice spatne = new Sachovnice(spatna); //konstruktor bez paramatru vygeneruje nahodnou sachovnici
         System.out.println("\nNahodne vygenerovana sachovnice:\n" + spatne);
         spatne.Vyres(false, 1000, 3); 
         System.out.println(separator);
         
       }
     
      /**
       * Moves the queens in their columns until it finds a nonconflict setting or overruns the maxCounter. Finally prints out nonconflict chessboard.
       * @param vypisy If true prints out chessboard after every step.
       * @param maxCounter Maximal number of steps after which computation is terminated.1
       * @param try How many times should it try to find the solution (Completely restart the computation).
       */ 
      public int Vyres(boolean vypisy, int maxCounter, int attempts)
      {
          Pozice[] damy = new Pozice[this.damy.length];
          int i = 0;
          int j = 0;
          for(i = 0; i<this.damy.length; i++)
            damy[i] = new Pozice(this.damy[i]);
          int[][] bitmap = new int[8][];
          for( i = 0; i<8; i++)
           {
               bitmap[i] = new int[8];
               System.arraycopy(this.bitmap[i],0,bitmap[i],0,this.bitmap.length);
            }
          int counter = 1;
          while ( counter <= attempts )
          {   if (Vyres(vypisy,maxCounter) == 0) 
                {
                    System.out.println("Vyreseno na " + counter + ". pokus");
                    return 0;
                }
              else
              {
                  System.out.println("Zkousim znova");
                  //nahrat znova pole damy a pole bitmap
                  for(i = 0; i<this.damy.length; i++)
                    this.damy[i] = new Pozice(damy[i]);
                  for(i = 0; i<8; i++)
                  {
                      System.arraycopy(bitmap[i],0,this.bitmap[i],0,bitmap.length);
                  }
                  counter++;
              }
          }
          System.out.println("Problem se nepodarilo vyresit ani po " + counter + " pokusech");
          return -1;
      }   
       
      /** 
       * Moves the queens in their columns until it finds a nonconflict setting or overruns the maxCounter. Finally prints out nonconflict chessboard.
       * @param vypisy If true prints out chessboard after every step.
       * @param maxCounter Maximal number of steps after which computation is terminated.
       */
      public int Vyres(boolean vypisy, int maxCounter)
      { 
       Pozice currentDama = null;
       int i = 0;
       int j = 0;
       int counter = 0;
       int konflikty = 0;
       Stav stav = null;
       SortedSet<Stav> stavy = new TreeSet<Stav>();
       int konfliktyCurrent = 0;
       while(counter != maxCounter) //dokud neni nalezen spravny stav, nebo prekroceni maxCounter ////uplne cele opakovat, dokud nebude stav, co mam hodnotu 0 nebo prekroceni maxCounter
       {   
           for(i = 0; i< damy.length; i++) //pro vsechny damy
           {
               currentDama = damy[i];
               konflikty = 0;
               bitmap[currentDama.getX()][currentDama.getY()] = 0; //vytahnout currentDamu z bitmapy, aby se nepocitala do konfliktuCelkem
               
               //spocitej konfliktyCelkem = konflikty ostatnich dam /2 
               for (j = 0; j < damy.length; j++)  //konflikty ostatnich dam
                  if ( j != i)
                   {
                       konflikty += konfliktu(damy[j]);  
                   }
               konflikty /= 2; //vsechny konflikty tam budou dvakrat
               for (j = 0; j < 8; j++) //pro kazde posunuti damy ve sloupci          
               {
                   if (j != currentDama.getY()) //krome stavu, ktery je prave aktivni, tj. zadne posunuti damy
                   {
                       //posun damu na radek j
                       konfliktyCurrent = konfliktu(currentDama.getX(), j); //spocitej konflikty damy[i] po posunuti 
                       stavy.add(new Stav(currentDama, konflikty + konfliktyCurrent, j));
                    }   
               }
               bitmap[currentDama.getX()][currentDama.getY()] = 1; //nakonec vratit currentDamu zpatky do bitmapy
           
           
           }
           stav = Sachovnice.getRandomLowest(stavy);   //vybrat podmnozinu se stavy s nejnizsimy hodnotami a nahodne jednu vybrat
           //prejit do tohoto stavu
           bitmap[stav.getDama().getX()][stav.getDama().getY()] = 0;
           stav.getDama().setY(stav.getKam());
           bitmap[stav.getDama().getX()][stav.getDama().getY()] = 1;
           stavy.clear(); //vymazat stavy
           counter++;
           if (vypisy)
             System.out.print("krok " + counter + ":\n" + this);
           if (stav.getKonfliktu() == 0)
             break;
       }  
       
       if (counter == maxCounter)
           {
               System.out.println("Ani po " + maxCounter + " krocich se nepodarilo vyresit problem");
               return -1;
            }
       else
         {
             System.out.println("Problem se podarilo vyresit po " + counter + " krocich."); 
             System.out.println("Vysledna sachovnice:\n" + this);
             return 0;
          }
       
     }
     
     
     
     /**
      * Returns one of the stav that has lowest number of conflicts (it chooses randomly among more items with same number of conflicts)
      * @return one of the stav that has lowest number of conflicts
      * @param stavy Sorted set of objects of type Stav.
      */
     private static Stav getRandomLowest(SortedSet<Stav> stavy) 
     {
        int nejnizsi = stavy.first().getKonfliktu();
        Iterator i = stavy.iterator();
        Stav last = null;
        while(i.hasNext())
        {
            last = (Stav)i.next();
            if (last.getKonfliktu() != nejnizsi)
              break;
        }
        SortedSet<Stav> nejnizsiStavy = stavy.headSet(last);  
        Stav chosen = null;
        Random mujRandom = new Random();
        chosen = (Stav)nejnizsiStavy.toArray()[mujRandom.nextInt(nejnizsiStavy.size() )];
        return chosen;
    }
         

}

   
